#!/usr/bin/python


class PrepareMessage(object):
    __slots__ = ("proposal_number")
    def __init__(self, sender, sequence):
        assert sender >= 0
        assert sequence >= 0
        self.proposal_number = (sequence, sender)

class PrepareResponse(object):
    """BUG: Need to send back the *accepted* proposal number for value."""
    __slots__ = ("proposal_number", "prepared", "value")
    def __init__(self, proposal, prepared, value):
        assert len(proposal) == 2
        assert prepared is True or prepared is False
        self.proposal_number = proposal
        self.prepared = prepared
        self.value = value

class AcceptMessage(object):
    __slots__ = ("proposal_number", "value")
    def __init__(self, sender, sequence, value):
        assert sender >= 0
        assert sequence >= 0
        assert value is not None
        self.proposal_number = (sequence, sender)
        self.value = value

class AcceptResponse(object):
    __slots__ = ("proposal_number", "accepted")
    def __init__(self, proposal, accepted):
        assert len(proposal) == 2
        assert accepted is True or accepted is False
        self.proposal_number = proposal
        self.accepted = accepted


class Acceptor(object):
    __slots__ = ("proposal_number", "value")

    def __init__(self):
        self.proposal_number = (0, -1)
        self.value = None

    def start(self, system, id):
        pass

    def messageReceived(self, system, source, data):
        assert source == data.proposal_number[1]

        # If the proposal number is more recent, use it
        accepted = False
        if data.proposal_number >= self.proposal_number:
            accepted = True
            self.proposal_number = data.proposal_number
            if isinstance(data, AcceptMessage):
                self.value = data.value

        # Disable NACKs if we didn't accept the value. Although they help more
        # complex implementations, it doesn't matter for the correctness of
        # the core Paxos algorithm.
        #~ if not accepted:
            #~ return

        response = None
        if isinstance(data, PrepareMessage):
            response = PrepareResponse(self.proposal_number, accepted, self.value)
        else:
            assert isinstance(data, AcceptMessage)
            if accepted:
                self.value = data.value
            response = AcceptResponse(self.proposal_number, accepted)
        system.sendMessage(source, response)


class Proposer(object):
    """Runs a single round of paxos to decide a value."""

    __slots__ = ("acceptor_ids", "proposal_value", "id", "sequence", "prepared", "accepted", "votes")

    def __init__(self, acceptor_ids, proposal_value):
        assert len(acceptor_ids) > 0
        assert proposal_value is not None

        self.acceptor_ids = acceptor_ids
        self.proposal_value = proposal_value
        self.id = None
        self.sequence = -1
        self.prepared = False
        self.accepted = False
        self.votes = set()

    def start(self, system, id):
        """Begin proposing a value."""
        assert self.id is None
        assert id >= 0

        self.id = id
        self.sequence += 1
        prepare = PrepareMessage(self.id, self.sequence)
        for acceptor_id in self.acceptor_ids:
            assert acceptor_id != self.id
            system.sendMessage(acceptor_id, prepare)

    def messageReceived(self, system, source, message):
        assert source in self.acceptor_ids

        # TODO: Count "reject" votes to determine we have "lost" sooner?
        accepted_vote = False
        if not self.prepared and isinstance(message, PrepareResponse):
            if message.proposal_number == (self.sequence, self.id) and message.prepared:
                accepted_vote = True
                if message.value is not None:
                    # The acceptor had a previously decided value: take it and propagate it
                    self.proposal_value = message.value
        elif (self.prepared and not self.accepted) and isinstance(message, AcceptResponse):
            if message.proposal_number == (self.sequence, self.id) and message.accepted:
                accepted_vote = True
        if not accepted_vote:
            return

        # Check if we have a majority of the votes
        self.votes.add(source)
        majority = len(self.acceptor_ids)/2 + 1
        if len(self.votes) < majority:
            return
        assert len(self.votes) == majority

        # Our proposal is accepted: send out the values
        if not self.prepared:
            self.prepared = True
            self.votes = set()
            accept = AcceptMessage(self.id, self.sequence, self.proposal_value)
            for acceptor_id in self.acceptor_ids:
                system.sendMessage(acceptor_id, accept)
        elif not self.accepted:
            self.accepted = True

